%%% -*- erlang-indent-level: 2 -*-
%%%-------------------------------------------------------------------
%%% File    : dialyzer_analysis.erl
%%% Author  : Tobias Lindahl <tobiasl@csd.uu.se>
%%% Description : Front end for the analysis employed by Dialyzer.
%%%
%%% Created : 7 May 2004 by Tobias Lindahl <tobiasl@csd.uu.se>
%%%-------------------------------------------------------------------

-module(dialyzer_analysis).
-export([start/3]).

-include("dialyzer.hrl").

%%%-------------------------------------------------------------------
%%% Defines below
%%%-------------------------------------------------------------------
 
-define(OPTS(Pid_), 
	[{type_warnings, {pid, Pid_}}, type_only, no_inline_fp,
	 {use_callgraph, fixpoint}, get_called_modules, 
	 {pmatch, no_duplicates}, {target, x86}]).

%%%-------------------------------------------------------------------
%%% Module-local record definitions below
%%%-------------------------------------------------------------------

-record(analysis_state, 
	{  
	  core_transform,
	  defines,
	  doc_plt,
	  fixpoint,
	  granularity,
	  include_dirs,
	  mcg=none,
	  opts,
	  parent,
	  plt,
	  start_from,
	  tmp_plt
	 }).

-record(state, {parent, legal_warnings}).

%%%-------------------------------------------------------------------

start(Parent, LegalWarnings, Analysis) ->
  State = #state{parent=Parent, legal_warnings=LegalWarnings},
  NewAnalysis1 = expand_files(Analysis),
  init_plt(NewAnalysis1),
  NewAnalysis2 = run_analysis(NewAnalysis1),
  loop(State, NewAnalysis2).

expand_files(Analysis) ->
  Files = Analysis#analysis.files,
  Ext = 
    case Analysis#analysis.start_from of
      byte_code -> ".beam";
      src_code -> ".erl"
    end,
  case expand_files(Files, Ext, []) of
    [] ->
      exit({error, "No files to analyze. Check analysis type."});
    NewFiles ->
      Analysis#analysis{files=NewFiles}
  end.

expand_files([File|Left], Ext, Acc) ->
  case filelib:is_dir(File) of
    true ->
      {ok, List} = file:list_dir(File),
      NewFiles = [filename:join(File, X)
		  || X <- List, filename:extension(X)==Ext],
      expand_files(Left, Ext, NewFiles++Acc);
    false ->
      expand_files(Left, Ext, [File|Acc])
  end;
expand_files([], _Ext, Acc) ->
  ordsets:from_list(Acc).

%% ----------------------------------------------------------------
%%
%%  Main Loop
%%

loop(State, Analysis = #analysis{}) ->
  AnalPid = Analysis#analysis.analysis_pid,
  Parent = State#state.parent,

  receive
    {AnalPid, log, LogMsg} ->
      send_log(Parent, LogMsg),
      loop(State, Analysis);
    {AnalPid, warnings, Warnings} ->
      case filter_warnings(State#state.legal_warnings, Warnings) of
	[] -> ok;
	SendWarnings ->
	  send_warnings(Parent, SendWarnings)
	end,
      loop(State, Analysis);
    {AnalPid, error, Msg} ->
      send_error(Parent, Msg),
      loop(State, Analysis);
    {AnalPid, module_callgraph, MCG} ->
      loop(State, Analysis#analysis{mcg={module_callgraph, MCG}});
    {AnalPid, done} ->
      NewAnalysis = analysis_done(Parent, Analysis),
      loop(State, NewAnalysis);
    {Parent, stop} ->
      exit(AnalPid, kill),
      ok;
    _Other ->
      %io:format("Received ~p\n", [Other]),
      loop(State, Analysis)
  end.

%% ----------------------------------------------------------------
%%
%%  Handle messages
%%

send_warnings(Parent, Warnings) ->
  Parent ! {self(), warnings, Warnings}.

filter_warnings(LegalWarnings, Warnings) ->
  [String || {Tag, String} <- Warnings, lists:member(Tag, LegalWarnings)].

send_analysis_done(Parent) ->
  Parent ! {self(), done}.


%% ----------------------------------------------------------------
%%
%%  Run the analysis
%%

run_analysis(Analysis) ->
  Self = self(),
  Fun = fun() -> analysis_start(Self, Analysis) end,
  Analysis#analysis{analysis_pid=spawn_link(Fun)}.

analysis_done(Parent, Analysis) ->
  case Analysis#analysis.fixpoint of
    first ->
      Msg = "\n========== Starting final run ==========\n",
      send_log(Parent, Msg),
      run_analysis(Analysis#analysis{fixpoint=second});
    second ->
      case Analysis#analysis.granularity == all of
	true ->
	  {module_callgraph, MCG} = Analysis#analysis.mcg,
	  case dialyzer_module_callgraph:external_calls(MCG) of
	    [] -> ok;
	    List -> 
	      InitPlt = Analysis#analysis.init_plt,
	      List1 = lists:filter(fun(X)->		
				       not dialyzer_plt:contains_mod(InitPlt, X)
				   end, List),
	      send_ext_calls(Parent, List1)
	  end;
	false ->
	  send_analysis_done(Parent)
      end,
      exit(normal);
    false ->
      send_analysis_done(Parent),
      exit(normal)
  end.

%%%-------------------------------------------------------------------

analysis_start(Parent, Analysis) ->     
  Opts1 = ?OPTS(self()),
  Opts2 =
    case Analysis#analysis.fixpoint of
      false ->
	[use_callgraph | Opts1];
      first ->
	[no_type_warnings|Opts1];
      second ->
	Opts1
    end,
  Opts3 =
    case Analysis#analysis.check_inline of
      true -> [check_for_inlining |Opts2];
      false -> Opts2
    end,
  Opts4 =
    case Analysis#analysis.granularity of
      module -> Opts3;
      all -> [{icode_type, {plt, Analysis#analysis.user_plt}}|Opts3]    
    end,

  State = #analysis_state{core_transform=Analysis#analysis.core_transform,
			  defines=Analysis#analysis.defines,
			  plt=Analysis#analysis.user_plt,
			  tmp_plt=dialyzer_plt:new(dialyzer_tmp_plt),
			  doc_plt=Analysis#analysis.doc_plt,
			  fixpoint=Analysis#analysis.fixpoint,
			  granularity=Analysis#analysis.granularity,
			  include_dirs=Analysis#analysis.include_dirs,
			  mcg=Analysis#analysis.mcg,
			  opts=Opts4,
			  parent=Parent,
			  start_from=Analysis#analysis.start_from
			 },

  case State#analysis_state.mcg of
    none ->
      analyze_all_files(Analysis#analysis.files, State);
    {module_callgraph, _MCG} ->
      analyze_callgraph(State)
  end,
  done(Parent).


%%============ Analyze files according to a callgraph ==============

analyze_callgraph(State=#analysis_state{mcg ={module_callgraph, Callgraph}}) ->
  analyze_sccs(Callgraph, State).

analyze_sccs(Callgraph, State) ->
  case dialyzer_module_callgraph:is_empty(Callgraph) of
    true ->
      ok;
    false ->
      case dialyzer_module_callgraph:take_first(Callgraph) of
	{[Mod], NewCallgraph} ->
	  %% Only one module
	  analyze_mod(Mod, State),
	  analyze_sccs(NewCallgraph, State);
	{Scc, NewCallgraph} ->
	  Parent = State#analysis_state.parent,
	  send_log(Parent, 
		   "-- Analyzing strongly connected component\n"
		   "-- (holding warnings)\n"),
	  Opts = State#analysis_state.opts,
	  TmpOpts = [no_type_warnings|Opts],
	  TmpState = State#analysis_state{opts=TmpOpts},
	  analyze_one_scc(Scc, TmpState),
	  send_log(Parent,
		   "-- Reached fixpoint for strongly connected component\n"
		   "-- Re-analyzing to get warnings\n"),
	  [analyze_mod(X, State) || X <- Scc],
	  analyze_sccs(NewCallgraph, State)
      end
  end.

analyze_one_scc(Scc, State) ->
  case analyze_one_scc(Scc, State, true) of
    fixpoint -> 
      ok;
    not_fixpoint -> 
      Msg = "-- Re-analyzing strongly connected component --\n",
      send_log(State#analysis_state.parent, Msg),
      analyze_one_scc(Scc, State)
  end.

analyze_one_scc([Mod|Left], State = #analysis_state{plt=Plt}, Fixpoint) ->
  Before = get_module_info(Plt, Mod),
  analyze_mod(Mod, State),
  After = get_module_info(Plt, Mod),
  case info_is_equal(Before, After) of
    true -> analyze_one_scc(Left, State, Fixpoint);
    false -> analyze_one_scc(Left, State, false)
  end;
analyze_one_scc([], _State, Fixpoint) ->
  case Fixpoint of
    true -> fixpoint;
    false -> not_fixpoint
  end.

info_is_equal([{MFA, Type1}|Left1], [{MFA, Type2}|Left2]) ->
  case erl_types:t_is_equal(Type1, Type2) of
    true ->
      info_is_equal(Left1, Left2);
    false ->
      false
  end;
info_is_equal([{MFA, RetType1, ArgTypes1}|Left1], 
	      [{MFA, RetType2, ArgTypes2}|Left2]) ->
  case lists:all(fun({X, Y}) ->erl_types:t_is_equal(X, Y)end,
		 lists:zip([RetType1|ArgTypes1], [RetType2|ArgTypes2])) of
    true ->
      info_is_equal(Left1, Left2);
    false ->
      false
  end;
info_is_equal([], []) ->
  true;
info_is_equal(_, _) ->
  false.


%%============ Analyze all files ==============

analyze_all_files(Files, State) ->
  ModMap = get_modules_by_dir(Files, State#analysis_state.start_from),
  case analyze_all_files(ModMap, State, []) of
    ok ->
      ok;
    {module_callgraph, NewMCG} ->
      send_module_callgraph(State#analysis_state.parent, NewMCG)
  end,
  ok.

analyze_all_files([{Mod, Dir}=File|Left],
 		  #analysis_state{start_from=byte_code,parent=Parent}=State,
 		  Acc) ->
  send_log(Parent, "Scanning " ++ Mod),
  ModPath = filename:join(Dir, Mod),
  case beam_lib:chunks(ModPath, [imports]) of
    {ok,{_,[{imports,Mfas}]}} ->
      Called = ordsets:from_list([M || {M,_,_} <- Mfas]),
      send_log(Parent, "...done\n"),
      analyze_all_files(Left, State, [{File, Called}|Acc]);
    _Other ->
      send_log(Parent, "...failed\n"),
      analyze_all_files(Left, State, Acc)
  end;
analyze_all_files([File|Left], State, Acc) ->
  case analyze_mod(File, State) of
    error ->
      analyze_all_files(Left, State, Acc);
    ok ->
      analyze_all_files(Left, State, Acc);
    {called_modules, Called} ->
      analyze_all_files(Left, State, [{File, Called}|Acc])
  end;
analyze_all_files([], _State, Acc) ->
  case Acc of
    [] ->
      ok;
    List when is_list(List) ->
      {module_callgraph, dialyzer_module_callgraph:construct(List)}
  end.

%%============ Analyze one module ==============

analyze_mod(Mod, State) ->
  Res =
    case State#analysis_state.granularity of
      module ->
	TmpPlt = mk_tmp_plt(State),
	TmpOpts = [{icode_type, {plt, TmpPlt}}|State#analysis_state.opts],
	TmpState = State#analysis_state{plt=TmpPlt, opts=TmpOpts},
	analyze_mod_1(Mod, TmpState);
      all ->
	analyze_mod_1(Mod, State)
    end,
  case Res of
    error -> 
      flush(),
      error;
    _ -> 
      insert_doc_info(Mod, State),
      Res
  end.

analyze_mod_1(Mod, State = #analysis_state{parent=Parent}) ->
  case analyze_mod_2(Mod, State) of
    {error, ErrorReport} ->
      send_error(Parent, ErrorReport),
      error;
    {ok, Msg} ->
      send_all_warnings(Parent),
      send_log(Parent, Msg),
      ok;
    {ok, Msg, CalledModules} ->
      send_all_warnings(Parent),
      send_log(Parent, Msg),
      {called_modules, CalledModules}
  end.

analyze_mod_2({Mod, Dir}, State = #analysis_state{parent=Parent,opts=Opts}) ->
  send_log(Parent, "Analyzing " ++ Mod),
  T0 = time_now(),
  Res =
    case State#analysis_state.start_from of
      src_code ->
	CompOpts = [no_inline,
		    return_errors, 
		    binary, to_core, 
		    {core_transform, State#analysis_state.core_transform}],
	DefaultIncl1 = ["..", "../incl", "../inc", "../include"],
	DefaultIncl2 = [{i, filename:join(Dir,X)} || X <- DefaultIncl1],
	UserIncludes = [{i, X} || X <- State#analysis_state.include_dirs],
	Includes = UserIncludes++DefaultIncl2,
	Defines = [{d, M, V} || {M,V} <- State#analysis_state.defines],
	ModPath = filename:join(Dir, Mod),
	case compile:file(ModPath, CompOpts++[{i, Dir}|Includes]++Defines) of
	  {ok, _, Core} ->
	    catch hipe:compile_core(list_to_atom(Mod), Core, [], Opts);
	  {error, Errors, _} ->
	    FormatedErrors = format_errors(Errors),
	    {error, lists:flatten(FormatedErrors)}
	end;
      byte_code ->
	analyze_beam_file(Dir, Mod, Opts, Parent, 
			  {core_transform, State#analysis_state.core_transform})
    end,
  %%garbage_collect(),
  case Res of
    {{ok, _}, {called_modules, C}} ->
      T = time_since(T0) / 1000,
      {ok, io_lib:format("...done in ~.2f secs\n",[T]), C};
    {ok, _} ->
      T = time_since(T0) / 1000,
      {ok, io_lib:format("...done in ~.2f secs\n",[T])};
    {error, What}  ->
      {error, 
       io_lib:format("\n*****In: ~s\n"
		     "*****Analysis failed with error report:\n~s\n",
		     [Mod, What])};
    {'EXIT', Why} ->
      {error, 
       io_lib:format("\n*****In: ~s\n"
		     "*****Analysis failed with error report:\n~w\n",
		     [Mod, Why])}
  end.

analyze_beam_file(Dir, ModName, Opts, Parent, CoreTransform) ->
  %% We must always set the code path, because the hipe compiler
  %% does a M:module_info() call!
  OldPath = code:get_path(),
  case code:add_patha(Dir) of
    true ->
      try
	Mod = list_to_atom(ModName),
	case beam_get_core(Dir, ModName, CoreTransform) of
	  error ->
	    send_log(Parent, " (byte code)"),
	    hipe:c(Mod, Opts);
	  Core ->
	    send_log(Parent, " (abstract code)"),
	    catch hipe:compile_core(Mod, Core, [], Opts)
	end
	after
	  %% Always restore the old code path here.
	  true = code:set_path(OldPath)
	end;
    false ->
      {error, "Could not add path: "++Dir}
  end.

%% beam_get_core(Dir, ModuleNameString) -> error | Core
%%  Retrieve the abstract code from the beam file and
%%  compile it to core. Returns 'error' if any problem occurs
%%  during this process.
%%
%%  XXX Potential problem: No inlining will be done, but
%%  the hipe compiler will look in the original beam file
%%  and will think that inlining occurred if the beam file
%%  was originally compiled with inlining.
beam_get_core(Dir, ModName, CoreTransform) ->
  File = filename:join(Dir, ModName),
  case beam_lib:chunks(File, [abstract_code]) of
    {ok,{_,[{abstract_code,{raw_abstract_v1,Abstr}}]}} ->
      CompOpts = [no_inline,to_core,CoreTransform],
      try compile:forms(Abstr, CompOpts) of
	  {ok,_,Core} -> Core;
	  _ ->
	  error
      catch
	error:_ ->
	  %% Compilation failed.
	  error
      end;
    _ ->
      %% No or unsuitable abstract code.
      error
  end.

format_errors([{Mod, Errors}|Left])->
  FormatedError = 
    [io_lib:format("~s:~w: ~s\n", [Mod, Line,apply(M,format_error, [Desc])])
     || {Line, M, Desc} <- Errors],
  [lists:flatten(FormatedError) | format_errors(Left)];
format_errors([]) ->
  [].
  
time_now() ->
  {T1, _} = statistics(runtime),
  {time_now,T1}.

time_since({time_now,T1}) ->
  {T2, _} = statistics(runtime),
  T = T2-T1,
  trunc(T).

%%============ Get module names ==============

get_modules_by_dir(Files, StartFrom)->
  get_modules_by_dir(Files, StartFrom, []).

get_modules_by_dir([File|Left], StartFrom, Acc)->
  BaseName = 
    case StartFrom of
      src_code -> filename:basename(File, ".erl");
      byte_code -> filename:basename(File, ".beam")
    end,
  Dir = filename:dirname(File),
  get_modules_by_dir(Left, StartFrom, [{BaseName, Dir}|Acc]);
get_modules_by_dir([], _StartFrom, Acc) ->
  lists:reverse(Acc).

%%============ Send and receive functions ==============

send_log(Parent, Msg) ->
  Parent ! {self(), log, Msg}.

send_error(Parent, Msg) ->
  Parent ! {self(), error, Msg}.

send_ext_calls(Parent, ExtCalls) ->
  Parent ! {self(), ext_calls, ExtCalls}.

send_all_warnings(Parent) ->
  case get_all_warnings() of
    [] -> ok;
    Warnings -> Parent ! {self(), warnings, Warnings}
  end.

get_all_warnings() ->
  receive
    {type_warning, TW} ->
      [TW | get_all_warnings()]
  after 0 ->
      []
  end.

send_module_callgraph(Parent, MCG) ->
  Parent ! {self(), module_callgraph, MCG}.

done(Parent) ->
  Parent ! {self(), done}.


flush() ->
  receive
    _ -> flush()
  after 0 ->
      ok
  end.


%%============ Handle the documentation table ==============

insert_doc_info(_, #analysis_state{doc_plt=none}) ->
  ok;
insert_doc_info({Mod, _Dir}, #analysis_state{plt=Plt, doc_plt=DocPlt}) ->
  TypeInfo = get_module_info(Plt, Mod),
  insert_doc_info_mfa_list(TypeInfo, DocPlt).

insert_doc_info_mfa_list([{MFA, ReturnType, ArgTypes}|Left], DocPlt) ->
  dialyzer_plt:insert(DocPlt, {MFA, ReturnType, ArgTypes}),
  insert_doc_info_mfa_list(Left, DocPlt);
insert_doc_info_mfa_list([], _DocPlt) ->
  ok.

%%============ Handle plt ==============


mk_tmp_plt(#analysis_state{plt=Plt, tmp_plt=TmpPlt}) ->
  dialyzer_plt:copy(Plt, TmpPlt),
  TmpPlt.

init_plt(#analysis{init_plt=InitPlt, user_plt=Plt, md5=MD5}) ->
  dialyzer_plt:copy(InitPlt, Plt),
  case MD5 of
    none -> ok;
    _ ->
      dialyzer_plt:insert(Plt, {md5, MD5}),
      ok
  end.

get_module_info(Plt, Mod0) ->
  Mod = case Mod0 of
	  {M, _Dir} -> list_to_atom(M);
	  M when is_list(M) -> list_to_atom(M)
	end,  
  Info = dialyzer_plt:find_mfa_info(Plt, Mod),
  lists:keysort(1, Info).
