%%%-------------------------------------------------------------------
%%% File    : dialyzer_plt.erl
%%% Author  : Tobias Lindahl <tobiasl@it.uu.se>
%%% Description : Interface to display information in the persistent 
%%%               lookup tables.
%%%
%%% Created : 23 Jul 2004 by Tobias Lindahl <tobiasl@it.uu.se>
%%%-------------------------------------------------------------------
-module(dialyzer_plt).

-include("dialyzer.hrl").	  %% file is automatically generated

-export([
	 check_init_plt/0,
	 copy/2,
	 contains_mod/2,
	 delete/1,
	 find_mfa_info/2,
	 from_dets/2,
	 insert/2,
	 new/1,
	 to_dets/2,
	 to_edoc/1,
	 to_edoc/4
	]).

new(Name) when is_atom(Name) ->
  ets:new(Name, [set, public]).

delete(Name) ->
  ets:delete(Name).

copy(From, To) ->
  List = ets:match_object(From, '_'),
  ets:delete_all_objects(To),
  ets:insert_new(To, List).

insert(Plt, Object) ->
  ets:insert(Plt, Object).

find_mfa_info(Plt, Mod) ->
  ets:match_object(Plt, {{Mod, '_', '_'}, '_', '_'}).

from_dets(Name, Dets) ->
  Plt = new(Name),
  {ok, D} = dets:open_file(Dets, [{access, read}]),
  true = ets:from_dets(Plt, D),
  ok = dets:close(D),
  Plt.

to_dets(Plt, Dets) ->
  MinSize = ets:info(Plt, size),
  {ok, Dets} = dets:open_file(Dets, [{min_no_slots, MinSize}]),
  ok = dets:from_ets(Dets, Plt),
  ok = dets:sync(Dets),
  ok = dets:close(Dets).

check_init_plt() ->
  MD5 = compute_md5(),
  case filelib:is_file(?INIT_PLT) andalso dets:is_dets_file(?INIT_PLT) of
    true ->
      case dets:open_file(?INIT_PLT, [{access, read}]) of
	{ok, Dets} ->
	  Res =
	    case dets:lookup(Dets, md5) of
	      {error, _What} -> {md5, MD5};
	      [] -> {md5, MD5};
	      [{md5, MD5}] -> ok;
	      [{md5, NewMD5}] -> {md5, NewMD5}
	    end,
	  ok = dets:close(Dets),
	  Res;
	{error, _Reason} ->
	  {md5, MD5}
      end;
    false -> {md5, MD5}
  end.      

compute_md5() ->
  LibDir = code:lib_dir("stdlib"),
  Dir = filename:join(LibDir, "ebin"),
  {ok,List} = file:list_dir(Dir),
  BeamFiles = [filename:join(Dir, X) 
	       || X<-List, filename:extension(X)==".beam"],
  Context = erlang:md5_init(),
  Fun = fun(X, Acc) ->
	    {ok,Bin} = file:read_file(X),
	    erlang:md5_update(Acc, Bin)
	end,
  FinalContext = lists:foldl(Fun, Context, BeamFiles),
  erlang:md5_final(FinalContext).

contains_mod(Plt, Mod0) ->
  Mod = 
    if is_list(Mod0) -> list_to_atom(Mod0);
       is_atom(Mod0) -> Mod0
    end,
  case ets:match_object(Plt, {{Mod, '_', '_'}, '_', '_'}) of
    [] -> false;
    _ -> true
  end.

to_edoc(PLT) ->
  to_edoc(PLT, '_', '_', '_').

to_edoc(PLT, M, F, A) ->
  List = ets:match_object(PLT, {{M, F, A}, '_', '_'}),
  SortedList = lists:keysort(1, List),
  lists:flatten(expand_edoc(SortedList, [])).

expand_edoc([{{M, F, A}, ReturnType, ArgTypes}| Left], M) ->
  case erl_types:t_is_any(ArgTypes) of
    true ->
      [io_lib:format("%% @spec ~w(~s) -> ~s\n", 
		     [F, expand_args_any(A), erl_types:t_to_string(ReturnType)])
       | expand_edoc(Left, M)];
    false ->
      [io_lib:format("%% @spec ~w(~s) -> ~s\n", 
		     [F, expand_args(ArgTypes), 
		      erl_types:t_to_string(ReturnType)])
       | expand_edoc(Left, M)]
  end;
expand_edoc(List = [{{M1, _F, _A}, _ReturnType, _ArgTypes}| _], _M) ->
  [io_lib:format("\n\n%% -------  Module: ~w -------\n\n", [M1]) | 
   expand_edoc(List, M1)];
expand_edoc([], _) ->
  [].

expand_args_any(0) ->
  [];
expand_args_any(1) ->
  ["_"];
expand_args_any(X) ->
  ["_,"|expand_args_any(X-1)].

expand_args([]) ->
  [];
expand_args([ArgType]) ->
  case erl_types:t_is_any(ArgType) of
    true -> ["_"];
    false -> [erl_types:t_to_string(ArgType)]
  end;
expand_args([ArgType|Left]) ->
  [case erl_types:t_is_any(ArgType) of
     true -> "_";
     false -> erl_types:t_to_string(ArgType)
   end ++
   ","|expand_args(Left)].
